from typing import Dict, List

API_KEY_ENV_KEY_MAP = {
    "base": "BASE_BLOCKSCOUT_API_KEY",
    "ethereum": "ETH_BLOCKSCOUT_API_KEY",
    "gnosis": "GNOSIS_BLOCKSCOUT_API_KEY",
    "optimism": "OPTIMISM_BLOCKSCOUT_API_KEY",
    "polygon": "POLYGON_BLOCKSCOUT_API_KEY",
}

NETWORKS: Dict[str, List[str]] = {
    "base": ["mainnet", "goerli", "sepolia"],
    "ethereum": ["mainnet", "goerli", "sepolia"],
    "gnosis": ["mainnet", "chiado"],
    "optimism": ["mainnet", "goerli", "sepolia"],
    "polygon": ["mainnet"],
}



import json
import os
from io import StringIO
from json import JSONDecodeError
from pathlib import Path
from tempfile import mkdtemp
from typing import IO, Any, Callable, Dict, Optional, Union
from unittest.mock import MagicMock

import _io  # type: ignore
import ape
import pytest
from ape.api import ExplorerAPI
from ape.exceptions import NetworkError
from ape.logging import logger
from ape.types import AddressType
from ape.utils import cached_property
from ape_solidity._utils import OUTPUT_SELECTION
from requests import Response

from ape_blockscout import Blockscout
from ape_blockscout.client import _APIClient
from ape_blockscout.types import BlockscoutResponse

ape.config.DATA_FOLDER = Path(mkdtemp()).resolve()
ape.config.PROJECT_FOLDER = Path(mkdtemp()).resolve()

MOCK_RESPONSES_PATH = Path(__file__).parent / "mock_responses"
CONTRACT_ADDRESS = "0xFe80e7afB7041c1592a2A5d8f617518c1591Aad4"
CONTRACT_ADDRESS_MAP = {
    "get_contract_response": CONTRACT_ADDRESS,
    "get_proxy_contract_response": "0x55A8a39bc9694714E2874c1ce77aa1E599461E18",
    "get_vyper_contract_response": "0xdA816459F1AB5631232FE5e97a05BBBb94970c95",
}
EXPECTED_ACCOUNT_TXNS_PARAMS = {
    "module": "account",
    "action": "txlist",
    "address": "0x1e59ce931B4CFea3fe4B875411e280e173cB7A9C",
    "endblock": None,
    "startblock": None,
    "offset": 100,
    "page": 1,
    "sort": "asc",
}
FOO_SOURCE_CODE = """
// SPDX-License-Identifier: AGPL-3.0
pragma solidity ^0.8.20;

import "@bar/bar.sol";

library MyLib {
    function insert(uint value) public returns (bool) {
        return true;
    }
}

contract foo {
    function register(uint value) public {
        require(MyLib.insert(value));
    }
}

contract fooWithConstructor {
    uint public value;
    constructor(uint _value) {
        value = _value;
    }
}
"""
BAR_SOURCE_CODE = r"""
// SPDX-License-Identifier: AGPL-3.0
pragma solidity ^0.8.20;

contract bar {
}
"""
APE_CONFIG_FILE = r"""
dependencies:
  - name: bar
    local: ./bar

solidity:
  import_remapping:
    - "@bar=bar"
"""


@pytest.fixture(scope="session")
def standard_input_json(library):
    return {
        "language": "Solidity",
        "sources": {
            "foo.sol": {"content": FOO_SOURCE_CODE},
            ".cache/bar/local/bar.sol": {"content": BAR_SOURCE_CODE},
        },
        "settings": {
            "optimizer": {"enabled": True, "runs": 200},
            "outputSelection": {
                ".cache/bar/local/bar.sol": {"": ["ast"], "*": OUTPUT_SELECTION},
                "subcontracts/foo.sol": {"": ["ast"], "*": OUTPUT_SELECTION},
            },
            "remappings": ["@bar=.cache/bar/local"],
            "viaIR": False,
        },
        "libraryname1": "MyLib",
        "libraryaddress1": library.address,
    }


@pytest.fixture(autouse=True)
def connection(explorer):
    with ape.networks.ethereum.mainnet.use_provider("infura") as provider:
        yield provider


def make_source(base_dir: Path, name: str, content: str):
    source_file = base_dir / f"{name}.sol"
    source_file.touch()
    source_file.write_text(content)


@pytest.fixture(scope="session", autouse=True)
def project():
    base_dir = ape.config.PROJECT_FOLDER
    contracts_dir = base_dir / "contracts"
    dependency_contracts_dir = base_dir / "bar" / "contracts"
    sub_contracts_dir = contracts_dir / "subcontracts"
    sub_contracts_dir.mkdir(exist_ok=True, parents=True)
    dependency_contracts_dir.mkdir(exist_ok=True, parents=True)

    make_source(sub_contracts_dir, "foo", FOO_SOURCE_CODE)
    make_source(dependency_contracts_dir, "bar", BAR_SOURCE_CODE)

    config_file = base_dir / "ape-config.yaml"
    config_file.unlink(missing_ok=True)
    config_file.write_text(APE_CONFIG_FILE)

    with ape.config.using_project(base_dir) as project:
        yield project


@pytest.fixture(scope="session")
def address():
    return CONTRACT_ADDRESS


@pytest.fixture(scope="session")
def account():
    return ape.accounts.test_accounts[0]


@pytest.fixture(scope="session")
def fake_connection():
    with ape.networks.ethereum.local.use_provider("test"):
        yield


@pytest.fixture
def no_api_key():
    key = os.environ.pop("BLOCKSCOUT_API_KEY", None)
    yield
    if key:
        os.environ["BLOCKSCOUT_API_KEY"] = key


@pytest.fixture
def explorer(get_explorer):
    return get_explorer("ethereum", "mainnet")


@pytest.fixture
def get_explorer(mocker):
    def fn(
        ecosystem_name: str = "ethereum",
        network_name: str = "development",
    ) -> ExplorerAPI:
        try:
            ecosystem = ape.networks.get_ecosystem(ecosystem_name)
        except NetworkError:
            # Use mock
            logger.warning(
                f"Ecosystem 'ape-{ecosystem_name}' not installed; resorting to a mock ecosystem."
            )
            ecosystem = mocker.MagicMock()
            ecosystem.name = ecosystem_name
            network = mocker.MagicMock()
            network.name = network_name
            network.ecosystem = ecosystem
            blockscout = ape.networks.get_ecosystem("ethereum").get_network("mainnet")
            explorer = Blockscout.model_construct(name=blockscout.name, network=network)
            network.explorer = explorer
            explorer.network = network
        else:
            network = ecosystem.get_network(network_name)
            explorer = network.explorer
            assert explorer is not None

        return explorer

    return fn


@pytest.fixture
def response(mocker):
    response = mocker.MagicMock(spec=Response)
    return BlockscoutResponse(response, "ethereum", raise_on_exceptions=False)


@pytest.fixture
def mock_backend(mocker):
    session = mocker.MagicMock()
    backend = MockBlockscoutBackend(mocker, session)
    _APIClient.session = session
    return backend


class MockBlockscoutBackend:
    def __init__(self, mocker, session):
        self._mocker = mocker
        self._session = session
        self._expected_base_uri = "https://eth.blockscout.com/api"  # Default
        self._handlers = {"get": {}, "post": {}}

    @cached_property
    def expected_uri_map(
        self,
    ) -> Dict[str, Dict[str, str]]:
        return {
            "base": {
                "mainnet": "https://base.blockscout.com/api",
                "goerli": "https://base-goerli.blockscout.com/api",
                "sepolia": "https://base-sepolia.blockscout.com/api",
            },
            "ethereum": {
                "mainnet": "https://eth.blockscout.com/api",
                "goerli": "https://eth-goerli.blockscout.com/api",
                "sepolia": "https://eth-sepolia.blockscout.com/api",
            },
            "gnosis": {
                "mainnet": "https://gnosis.blockscout.com/api",
                "chiado": "https://gnosis-chiado.blockscout.com/api",
            },
            "optimism": {
                "mainnet": "https://optimism.blockscout.com/api",
                "goerli": "https://optimism-goerli.blockscout.com/api",
                "sepolia": "https://optimism-sepolia.blockscout.com/api",
            },
            "polygon": {
                "mainnet": "https://polygon.blockscout.com/api",
            },
        }

    def set_network(self, ecosystem: str, network: str):
        self._expected_base_uri = self.expected_uri_map[ecosystem][network.replace("-fork", "")]

    def add_handler(
        self,
        method: str,
        module: str,
        expected_params: Dict,
        return_value: Optional[Any] = None,
        side_effect: Optional[Callable] = None,
    ):
        if isinstance(return_value, (str, dict)):
            return_value = self.get_mock_response(return_value)

        def handler(self, method, base_uri, params=None, data=None, headers=None):
            actual_params = params if method.lower() == "get" else data
            for key, val in expected_params.items():
                if key not in expected_params:
                    # Allow skipping certain assertions
                    continue

                # Handler StringIO objects
                if isinstance(val, StringIO):
                    assert isinstance(actual_params[key], StringIO)
                    text = actual_params[key].read()
                    if text:
                        try:
                            actual_json = json.loads(text)
                        except JSONDecodeError:
                            pytest.fail(f"Response text is not JSON: '{text}'.")
                            return
                    else:
                        # Empty.
                        actual_json = {}

                    val = val.read()
                    expected_json = json.loads(val) if val else {}
                    assert actual_json == expected_json

                else:
                    msg = f"expected={key}"
                    if params:
                        msg = f"{msg} module={params['module']} action={params['action']}"

                    assert actual_params[key] == val, msg

            if return_value:
                return return_value

            elif side_effect:
                result = side_effect()
                return result if isinstance(result, Response) else self.get_mock_response(result)

        self._handlers[method.lower()][module] = handler
        self._session.request.side_effect = self.handle_request

    def handle_request(self, method, base_uri, timeout, headers=None, params=None, data=None):
        if params and "apikey" in params:
            del params["apikey"]
        if data and "apiKey" in data:
            del data["apiKey"]

        assert base_uri == self._expected_base_uri

        if params:
            module = params.get("module")
        elif data:
            module = data.get("module")
        else:
            raise AssertionError("Expected either 'params' or 'data'.")

        handler = self._handlers[method.lower()][module]
        return handler(self, method, base_uri, headers=headers, params=params, data=data)

    def setup_mock_get_contract_type_response(self, file_name: str):
        response = self._get_contract_type_response(file_name)
        address = CONTRACT_ADDRESS_MAP[file_name]
        expected_params = self._expected_get_ct_params(address)
        self.add_handler("GET", "contract", expected_params, return_value=response)
        response.expected_address = address
        return response

    def setup_mock_get_contract_type_response_with_throttling(
        self, file_name: str, retries: int = 2
    ):
        response = self._get_contract_type_response(file_name)
        address = CONTRACT_ADDRESS_MAP[file_name]
        expected_params = self._expected_get_ct_params(address)
        throttled = self._mocker.MagicMock(spec=Response)
        throttled.status_code = 429

        class ThrottleMock:
            counter = 0

            def side_effect(self):
                if self.counter < retries:
                    self.counter += 1
                    return throttled

                return response

        throttler = ThrottleMock()
        self.add_handler("GET", "contract", expected_params, side_effect=throttler.side_effect)
        response.expected_address = address
        return throttler, response

    def _get_contract_type_response(self, file_name: str) -> Any:
        test_data_path = MOCK_RESPONSES_PATH / f"{file_name}.json"
        with open(test_data_path) as response_data_file:
            return self.get_mock_response(response_data_file, file_name=file_name)

    def _expected_get_ct_params(self, address: str) -> Dict:
        return {"module": "contract", "action": "getsourcecode", "address": address}

    def setup_mock_account_transactions_response(self, address: Optional[AddressType] = None):
        file_name = "get_account_transactions.json"
        test_data_path = MOCK_RESPONSES_PATH / file_name

        if address:
            params = EXPECTED_ACCOUNT_TXNS_PARAMS.copy()
            params["address"] = address
        else:
            params = EXPECTED_ACCOUNT_TXNS_PARAMS

        with open(test_data_path) as response_data_file:
            response = self.get_mock_response(response_data_file, file_name=file_name)
            self.add_handler("GET", "account", params, return_value=response)
            self.set_network("ethereum", "mainnet")
            return response

    def setup_mock_account_transactions_with_ctor_args_response(
        self, address: Optional[AddressType] = None
    ):
        file_name = "get_account_transactions_with_ctor_args.json"
        test_data_path = MOCK_RESPONSES_PATH / file_name

        if address:
            params = EXPECTED_ACCOUNT_TXNS_PARAMS.copy()
            params["address"] = address
        else:
            params = EXPECTED_ACCOUNT_TXNS_PARAMS

        with open(test_data_path) as response_data_file:
            response = self.get_mock_response(response_data_file, file_name=file_name)
            self.add_handler("GET", "account", params, return_value=response)
            self.set_network("ethereum", "mainnet")
            return response

    def get_mock_response(
        self, response_data: Optional[Union[IO, Dict, str, MagicMock]] = None, **kwargs
    ):
        if isinstance(response_data, str):
            return self.get_mock_response({"result": response_data})

        elif isinstance(response_data, _io.TextIOWrapper):
            return self.get_mock_response(json.load(response_data), **kwargs)

        elif isinstance(response_data, MagicMock):
            # Mock wasn't set.
            response_data = {}

        response = self._mocker.MagicMock(spec=Response)
        response.json.return_value = response_data
        response.text = json.dumps(response_data or {})

        response.status_code = 200

        for key, val in kwargs.items():
            setattr(response, key, val)

        return response


@pytest.fixture
def verification_params(address_to_verify, standard_input_json):
    ctor_args = ""  # noqa: E501

    return {
        "action": "verifysourcecode",
        "codeformat": "solidity-standard-json-input",
        "constructorArguments": ctor_args,
        "contractaddress": address_to_verify,
        "contractname": "foo.sol:foo",
        "evmversion": None,
        "licenseType": 1,
        "module": "contract",
        "optimizationUsed": 1,
        "runs": 200,
        "sourceCode": StringIO(json.dumps(standard_input_json)),
    }


@pytest.fixture(scope="session")
def verification_params_with_ctor_args(
    address_to_verify_with_ctor_args, library, standard_input_json
):
    # abi-encoded representation of uint256 value 42
    ctor_args = "000000000000000000000000000000000000000000000000000000000000002a"  # noqa: E501

    json_data = standard_input_json.copy()
    json_data["libraryaddress1"] = "0xF2Df0b975c0C9eFa2f8CA0491C2d1685104d2488"

    return {
        "action": "verifysourcecode",
        "codeformat": "solidity-standard-json-input",
        "constructorArguments": ctor_args,
        "contractaddress": address_to_verify_with_ctor_args,
        "contractname": "foo.sol:fooWithConstructor",
        "evmversion": None,
        "licenseType": 1,
        "module": "contract",
        "optimizationUsed": 1,
        "runs": 200,
        "sourceCode": StringIO(json.dumps(json_data)),
    }


@pytest.fixture(scope="session")
def chain():
    return ape.chain


@pytest.fixture(scope="session")
def solidity(project):
    return project.compiler_manager.solidity


@pytest.fixture(scope="session")
def library(account, project, chain, solidity):
    lib = account.deploy(project.MyLib)
    chain.contracts._local_contract_types[lib.address] = lib.contract_type
    solidity.add_library(lib)
    return lib


@pytest.fixture(scope="session")
def address_to_verify(fake_connection, library, project, account):
    _ = library  # Ensure library is deployed first.
    foo = project.foo.deploy(sender=account)
    ape.chain.contracts._local_contract_types[address] = foo.contract_type
    return foo.address


@pytest.fixture(scope="session")
def address_to_verify_with_ctor_args(fake_connection, project, account):
    # Deploy the library first.
    library = account.deploy(project.MyLib)
    ape.chain.contracts._local_contract_types[library.address] = library.contract_type

    # Add the library to recompile contract `foo`.
    solidity = project.compiler_manager.solidity
    solidity.add_library(library)

    foo = project.fooWithConstructor.deploy(42, sender=account)
    ape.chain.contracts._local_contract_types[address] = foo.contract_type
    return foo.address


@pytest.fixture(scope="session")
def expected_verification_log(address_to_verify):
    return (
        "Contract verification successful!\n"
        f"https://eth.blockscout.io/address/{address_to_verify}#code"
    )


@pytest.fixture(scope="session")
def expected_verification_log_with_ctor_args(address_to_verify_with_ctor_args):
    return (
        "Contract verification successful!\n"
        f"https://eth.blockscout.io/address/{address_to_verify_with_ctor_args}#code"
    )



from ape import plugins

from .config import BlockscoutConfig
from .explorer import Blockscout
from .utils import NETWORKS


@plugins.register(plugins.ExplorerPlugin)
def explorers():
    for ecosystem_name in NETWORKS:
        for network_name in NETWORKS[ecosystem_name]:
            yield ecosystem_name, network_name, Blockscout
            yield ecosystem_name, f"{network_name}-fork", Blockscout


@plugins.register(plugins.Config)
def config_class():
    return BlockscoutConfig



from ape.api.config import PluginConfig


class EcosystemConfig(PluginConfig):
    rate_limit: int = 5  # Requests per second
    retries: int = 5  # Number of retries before giving up


class BlockscoutConfig(PluginConfig):
    ethereum: EcosystemConfig = EcosystemConfig()
    gnosis: EcosystemConfig = EcosystemConfig()
    polygon: EcosystemConfig = EcosystemConfig()
    optimism: EcosystemConfig = EcosystemConfig()



import pytest

from ape_blockscout import NETWORKS
from ape_blockscout.exceptions import BlockscoutTooManyRequestsError

# A map of each mock response to its contract name for testing `get_contract_type()`.
EXPECTED_CONTRACT_NAME_MAP = {
    "get_contract_response": "BoredApeYachtClub",
    "get_proxy_contract_response": "MIM-UST-f",
    "get_vyper_contract_response": "yvDAI",
}
TRANSACTION = "0x0da22730986e96aaaf5cedd5082fea9fd82269e41b0ee020d966aa9de491d2e6"
PUBLISH_GUID = "123"

# Every supported ecosystem / network combo as `[("ecosystem", "network") ... ]`
ecosystems_and_networks = [
    p
    for plist in [
        [(e, n) for n in nets] + [(e, f"{n}-fork") for n in nets] for e, nets in NETWORKS.items()
    ]
    for p in plist
]
base_url_test = pytest.mark.parametrize(
    "ecosystem,network,url",
    [
        # Base
        ("base", "mainnet", "base.blockscout.com"),
        ("base", "mainnet-fork", "base.blockscout.com"),
        # Ethereum
        ("ethereum", "mainnet", "eth.blockscout.com"),
        ("ethereum", "mainnet-fork", "eth.blockscout.com"),
        ("ethereum", "goerli", "eth-goerli.blockscout.com"),
        ("ethereum", "goerli-fork", "eth-goerli.blockscout.com"),
        ("ethereum", "sepolia", "eth-sepolia.blockscout.com"),
        ("ethereum", "sepolia-fork", "eth-sepolia.blockscout.com"),
        # Gnosis
        ("gnosis", "mainnet", "gnosis.blockscout.com"),
        ("gnosis", "mainnet-fork", "gnosis.blockscout.com"),
        ("gnosis", "chiado", "gnosis-chiado.blockscout.com"),
        ("gnosis", "chiado-fork", "gnosis-chiado.blockscout.com"),
        # Optimism
        ("optimism", "mainnet", "optimism.blockscout.com"),
        ("optimism", "mainnet-fork", "optimism.blockscout.com"),
        ("optimism", "goerli", "optimism-goerli.blockscout.com"),
        ("optimism", "goerli-fork", "optimism-goerli.blockscout.com"),
        ("optimism", "sepolia", "optimism-sepolia.blockscout.com"),
        ("optimism", "sepolia-fork", "optimism-sepolia.blockscout.com"),
        # Polygon
        ("polygon", "mainnet", "polygon.blockscout.com"),
        ("polygon", "mainnet-fork", "polygon.blockscout.com"),
    ],
)


@base_url_test
def test_get_address_url(ecosystem, network, url, address, get_explorer):
    expected = f"https://{url}/address/{address}"
    explorer = get_explorer(ecosystem, network)
    actual = explorer.get_address_url(address)
    assert actual == expected


@base_url_test
def test_get_transaction_url(ecosystem, network, url, get_explorer):
    expected = f"https://{url}/tx/{TRANSACTION}"
    explorer = get_explorer(ecosystem, network)
    actual = explorer.get_transaction_url(TRANSACTION)
    assert actual == expected


@pytest.mark.parametrize("ecosystem,network", ecosystems_and_networks)
def test_get_contract_type_ecosystems_and_networks(
    mock_backend,
    ecosystem,
    network,
    get_explorer,
):
    # This test parametrizes getting contract types across ecosystem / network combos
    mock_backend.set_network(ecosystem, network)
    response = mock_backend.setup_mock_get_contract_type_response("get_contract_response")
    explorer = get_explorer(ecosystem, network)
    actual = explorer.get_contract_type(response.expected_address)
    contract_type_from_lowered_address = explorer.get_contract_type(
        response.expected_address.lower()
    )
    assert actual is not None
    assert actual == contract_type_from_lowered_address

    actual = actual.name
    expected = EXPECTED_CONTRACT_NAME_MAP[response.file_name]
    assert actual == expected


@pytest.mark.parametrize(
    "file_name", ("get_proxy_contract_response", ("get_vyper_contract_response"))
)
def test_get_contract_type_additional_types(mock_backend, file_name, explorer):
    # This test parametrizes getting edge-case contract types.
    # NOTE: Purposely not merged with test above to avoid adding a new dimension
    #  to the parametrization.
    response = mock_backend.setup_mock_get_contract_type_response(file_name)
    actual = explorer.get_contract_type(response.expected_address).name
    expected = EXPECTED_CONTRACT_NAME_MAP[response.file_name]
    assert actual == expected


def test_get_contract_type_with_rate_limiting(mock_backend, explorer):
    """
    This test ensures the rate limiting logic in the Explorer client works.
    """

    file_name = "get_vyper_contract_response"
    setter_upper = mock_backend.setup_mock_get_contract_type_response_with_throttling
    throttler, response = setter_upper(file_name)

    # We still eventually get the response.
    actual = explorer.get_contract_type(response.expected_address).name
    expected = EXPECTED_CONTRACT_NAME_MAP[response.file_name]
    assert actual == expected
    assert throttler.counter == 2  # Prove that it actually throttled.


def test_too_many_requests_error(no_api_key, response):
    actual = str(BlockscoutTooManyRequestsError(response, "ethereum"))
    assert "BLOCKSCOUT_API_KEY" in actual



import os
from typing import TYPE_CHECKING, Union

from ape.exceptions import ApeException
from requests import Response

from ape_blockscout.utils import API_KEY_ENV_KEY_MAP

if TYPE_CHECKING:
    from ape_blockscout.types import BlockscoutResponse, ResponseValue


class BlockscoutException(ApeException):
    """
    A base exception in the ape-blockscout plugin.
    """


class UnsupportedEcosystemError(BlockscoutException):
    """
    Raised when there is no Blockscout buildout for ecosystem.
    """

    def __init__(self, ecosystem: str):
        super().__init__(f"Unsupported Ecosystem: {ecosystem}")


class UnsupportedNetworkError(BlockscoutException):
    """
    Raised when there is no Blockscout buildout for ecosystem.
    """

    def __init__(self, ecosystem_name: str, network_name: str):
        super().__init__(f"Unsupported Network for Ecosystem '{ecosystem_name}': {network_name}")


class BlockscoutResponseError(BlockscoutException):
    """
    Raised when the response is not correct.
    """

    def __init__(self, response: Union[Response, "BlockscoutResponse"], message: str):
        if not isinstance(response, Response):
            response = response.response

        self.response = response
        super().__init__(f"Response indicated failure: {message}")


class UnhandledResultError(BlockscoutResponseError):
    """
    Raised in specific client module where the result from Blockscout
    has an unhandled form.
    """

    def __init__(self, response: Union[Response, "BlockscoutResponse"], value: "ResponseValue"):
        message = f"Unhandled response format: {value}"
        super().__init__(response, message)


class BlockscoutTooManyRequestsError(BlockscoutResponseError):
    """
    Raised after being rate-limited by Blockscout.
    """

    def __init__(self, response: Union[Response, "BlockscoutResponse"], ecosystem: str):
        message = "Blockscout API server rate limit exceeded."
        api_key_name = API_KEY_ENV_KEY_MAP[ecosystem]
        if not os.environ.get(api_key_name):
            message = f"{message}. Try setting {api_key_name}'."

        super().__init__(response, message)


def get_request_error(response: Response, ecosystem: str) -> BlockscoutResponseError:
    response_data = response.json()
    if "result" in response_data and response_data["result"]:
        message = response_data["result"]
    elif "message" in response_data:
        message = response_data["message"]
    else:
        message = response.text

    if "max rate limit reached" in response.text.lower():
        return BlockscoutTooManyRequestsError(response, ecosystem)

    return BlockscoutResponseError(response, message)



import json
import os
import random
import time
from io import StringIO
from typing import Dict, Iterator, List, Optional

from ape.logging import logger
from ape.utils import USER_AGENT, ManagerAccessMixin
from requests import Session
from yarl import URL

from ape_blockscout.exceptions import (
    UnhandledResultError,
    UnsupportedEcosystemError,
    UnsupportedNetworkError,
)
from ape_blockscout.types import BlockscoutResponse, ContractCreationResponse, SourceCodeResponse
from ape_blockscout.utils import API_KEY_ENV_KEY_MAP, NETWORKS


def get_blockscout_uri(ecosystem_name: str, network_name: str) -> str:
    if ecosystem_name not in API_KEY_ENV_KEY_MAP.keys():
        raise UnsupportedEcosystemError(ecosystem_name)

    if network_name not in NETWORKS[ecosystem_name]:
        raise UnsupportedNetworkError(ecosystem_name, network_name)

    # Blockscout uses "eth" for ethereum, but we use "ethereum" in ape.
    ecosystem_name = "eth" if ecosystem_name == "ethereum" else ecosystem_name

    path_name = ecosystem_name if network_name == "mainnet" else f"{ecosystem_name}-{network_name}"

    return f"https://{path_name}.blockscout.com"


class _APIClient(ManagerAccessMixin):
    DEFAULT_HEADERS = {"User-Agent": USER_AGENT}
    session: Session = Session()
    ecosystem_name: str

    def __init__(self, ecosystem_name: str, network_name: str, module_name: str):
        self._ecosystem_name = ecosystem_name
        self._network_name = network_name
        self._module_name = module_name
        self._last_call = 0.0

    @property
    def base_uri(self) -> str:
        return get_blockscout_uri(self._ecosystem_name, self._network_name) + "/api"

    @property
    def base_params(self) -> Dict:
        return {"module": self._module_name}

    @property
    def _rate_limit(self) -> int:
        config = self.config_manager.get_config("blockscout")
        return getattr(config, self.network_manager.ecosystem.name.lower()).rate_limit

    @property
    def _retries(self) -> int:
        config = self.config_manager.get_config("blockscout")
        return getattr(config, self.network_manager.ecosystem.name.lower()).retries

    @property
    def _min_time_between_calls(self) -> float:
        return 1 / self._rate_limit  # seconds / calls per second

    @property
    def _clean_uri(self) -> str:
        url: URL = URL(self.base_uri).with_user(None).with_password(None)
        return f"{url.with_path('')}/[hidden]" if url.path else f"{url}"

    def _get(
        self,
        params: Optional[Dict] = None,
        headers: Optional[Dict[str, str]] = None,
        raise_on_exceptions: bool = True,
    ) -> BlockscoutResponse:
        params = self.__authorize(params)

        # Rate limit
        if time.time() - self._last_call < self._min_time_between_calls:
            time_to_sleep = self._min_time_between_calls - (time.time() - self._last_call)
            logger.debug(f"Sleeping {time_to_sleep} seconds to avoid rate limit")
            # NOTE: Sleep time is in seconds (float for subseconds)
            time.sleep(time_to_sleep)

        self._last_call = time.time()

        return self._request(
            "GET",
            params=params,
            headers=headers,
            raise_on_exceptions=raise_on_exceptions,
        )

    def _post(
        self, json_dict: Optional[Dict] = None, headers: Optional[Dict[str, str]] = None
    ) -> BlockscoutResponse:
        data = self.__authorize(json_dict)
        return self._request("POST", data=data, headers=headers)

    def _request(
        self,
        method: str,
        raise_on_exceptions: bool = True,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
        data: Optional[Dict] = None,
    ) -> BlockscoutResponse:
        headers = headers or self.DEFAULT_HEADERS
        for i in range(self._retries):
            logger.debug(f"Request sent to {self._clean_uri}.")
            response = self.session.request(
                method.upper(),
                self.base_uri,
                headers=headers,
                params=params,
                data=data,
                timeout=1024,
            )
            if response.status_code == 429:
                time_to_sleep = 2**i
                logger.debug(f"Request was throttled. Retrying in {time_to_sleep} seconds.")
                time.sleep(time_to_sleep)
                continue

            # Recieved a real response unrelated to rate limiting.
            if raise_on_exceptions:
                response.raise_for_status()
            elif not 200 <= response.status_code < 300:
                logger.error(f"Response was not successful: {response.text}")

            break

        return BlockscoutResponse(response, self.ecosystem_name, raise_on_exceptions)

    def __authorize(self, params_or_data: Optional[Dict] = None) -> Optional[Dict]:
        env_var_key = API_KEY_ENV_KEY_MAP.get(self.ecosystem_name)
        if not env_var_key:
            return params_or_data

        api_key = os.environ.get(env_var_key)
        if api_key and (not params_or_data or "apikey" not in params_or_data):
            params_or_data = params_or_data or {}
            api_key = random.choice(api_key.split(","))
            params_or_data["apikey"] = api_key.strip()

        return params_or_data


class ContractClient(_APIClient):
    def __init__(self, ecosystem_name: str, network_name: str, address: str):
        self._address = address
        self.ecosystem_name = ecosystem_name

        super().__init__(
            ecosystem_name=ecosystem_name, network_name=network_name, module_name="contract"
        )

    def get_source_code(self) -> SourceCodeResponse:
        params = {
            **self.base_params,
            "action": "getsourcecode",
            "address": self._address,
        }

        result = self._get(params=params)

        if not (result_list := result.value or []):
            return SourceCodeResponse()

        elif len(result_list) > 1:
            raise UnhandledResultError(result, result_list)

        data = result_list[0]
        if not isinstance(data, dict):
            raise UnhandledResultError(result, data)

        abi = data.get("ABI") or ""
        name = data.get("ContractName") or "unknown"
        return SourceCodeResponse(abi, name)

    def verify_source_code(
        self,
        standard_json_output: Dict,
        compiler_version: str,
        contract_name: Optional[str] = None,
        optimization_used: bool = False,
        optimization_runs: Optional[int] = 200,
        constructor_arguments: Optional[str] = None,
        evm_version: Optional[str] = None,
        license_type: Optional[int] = None,
        libraries: Optional[Dict[str, str]] = None,
    ) -> str:
        libraries = libraries or {}
        if len(libraries) > 10:
            raise ValueError(f"Can only have up to 10 libraries (received {len(libraries)}).")

        if not compiler_version.startswith("v"):
            compiler_version = f"v{compiler_version}"

        if "sourceCode" in standard_json_output:
            source_code = standard_json_output["sourceCode"]
            code_format = "solidity-single-file"
        else:
            source_code = StringIO(json.dumps(standard_json_output))
            code_format = "solidity-standard-json-input"

        json_dict = {
            **self.base_params,
            "action": "verifysourcecode",
            "codeformat": code_format,
            "compilerversion": compiler_version,
            "constructorArguments": constructor_arguments,
            "contractaddress": self._address,
            "contractname": contract_name,
            "evmversion": evm_version,
            "licenseType": license_type,
            "optimizationUsed": int(optimization_used),
            "runs": optimization_runs,
            "sourceCode": source_code,
        }
        iterator = 1
        for lib_address, lib_name in libraries.items():
            json_dict[f"libraryname{iterator}"] = lib_name
            json_dict[f"libraryaddress{iterator}"] = lib_address
            iterator += 1

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        return str(self._post(json_dict=json_dict, headers=headers).value)

    def check_verify_status(self, guid: str) -> str:
        json_dict = {**self.base_params, "action": "checkverifystatus", "guid": guid}
        response = self._get(params=json_dict, raise_on_exceptions=False)
        return str(response.value)

    def get_creation_data(self) -> List[ContractCreationResponse]:
        params = {
            **self.base_params,
            "action": "getcontractcreation",
            "contractaddresses": [self._address],
        }
        result = self._get(params=params)
        items = result.value or []
        if not isinstance(items, list):
            raise ValueError("Expecting list.")

        return [ContractCreationResponse(**item) for item in items]


class AccountClient(_APIClient):
    def __init__(self, ecosystem_name: str, network_name: str, address: str):
        self._address = address
        self.ecosystem_name = ecosystem_name

        super().__init__(
            ecosystem_name=ecosystem_name, network_name=network_name, module_name="account"
        )

    def get_all_normal_transactions(
        self,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        offset: int = 100,
        sort: str = "asc",
    ) -> Iterator[Dict]:
        page_num = 1
        last_page_results = offset  # Start at offset to trigger iteration
        while last_page_results == offset:
            page = self._get_page_of_normal_transactions(
                page_num, start_block, end_block, offset=offset, sort=sort
            )

            if len(page):
                yield from page

            last_page_results = len(page)
            page_num += 1

    def _get_page_of_normal_transactions(
        self,
        page: int,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        offset: int = 100,
        sort: str = "asc",
    ) -> List[Dict]:
        params = {
            **self.base_params,
            "action": "txlist",
            "address": self._address,
            "startblock": start_block,
            "endblock": end_block,
            "page": page,
            "offset": offset,
            "sort": sort,
        }
        result = self._get(params=params)

        if not isinstance(result.value, list):
            raise UnhandledResultError(result, result.value)

        return result.value


class ClientFactory:
    def __init__(self, ecosystem_name: str, network_name: str):
        self._ecosystem_name = ecosystem_name
        self._network_name = network_name

    def get_contract_client(self, contract_address: str) -> ContractClient:
        return ContractClient(
            ecosystem_name=self._ecosystem_name,
            network_name=self._network_name,
            address=contract_address,
        )

    def get_account_client(self, account_address: str) -> AccountClient:
        return AccountClient(
            ecosystem_name=self._ecosystem_name,
            network_name=self._network_name,
            address=account_address,
        )



import json
from json.decoder import JSONDecodeError
from typing import Optional

from ape.api import ExplorerAPI
from ape.contracts import ContractInstance
from ape.exceptions import APINotImplementedError, ProviderNotConnectedError
from ape.logging import logger
from ape.types import AddressType, ContractType

from ape_blockscout.client import ClientFactory, get_blockscout_uri


class Blockscout(ExplorerAPI):
    def get_address_url(self, address: str) -> str:
        ecosystem_uri = get_blockscout_uri(
            ecosystem_name=self.network.ecosystem.name,
            network_name=self.network.name.replace("-fork", ""),
        )

        return f"{ecosystem_uri}/address/{address}"

    def get_transaction_url(self, transaction_hash: str) -> str:
        ecosystem_uri = get_blockscout_uri(
            ecosystem_name=self.network.ecosystem.name,
            network_name=self.network.name.replace("-fork", ""),
        )

        return f"{ecosystem_uri}/tx/{transaction_hash}"

    @property
    def _client_factory(self) -> ClientFactory:
        return ClientFactory(
            ecosystem_name=self.network.ecosystem.name,
            network_name=self.network.name.replace("-fork", ""),
        )

    def get_contract_type(self, address: AddressType) -> Optional[ContractType]:
        if not self.conversion_manager.is_type(address, AddressType):
            # Handle non-checksummed addresses
            address = self.conversion_manager.convert(str(address), AddressType)

        client = self._client_factory.get_contract_client(address)

        source_code = client.get_source_code()
        if not (abi_string := source_code.abi):
            return None

        try:
            abi = json.loads(abi_string)
        except JSONDecodeError as err:
            logger.error(f"Error with contract ABI: {err}")
            return None

        contract_type = ContractType(abi=abi, contractName=source_code.name)
        if source_code.name == "Vyper_contract" and "symbol" in contract_type.view_methods:
            try:
                contract = ContractInstance(address, contract_type)
                contract_type.name = contract.symbol() or contract_type.name
            except ProviderNotConnectedError:
                pass

        return contract_type

    def publish_contract(self, address: AddressType):
        raise APINotImplementedError("Publishing contracts is not supported by Blockscout")
